//=======================================================================
// mrftv.h : The general implementation (ND) of Markov Random Field 
//           Tensor Voting (MRFTV)
//
// Copyright (c) 2009 Tai-Pang Wu (pang@cse.ust.hk) and HKUST x CUHK
// 
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//=======================================================================


#ifndef _MRFTV_H_
#define _MRFTV_H_

#include "StdAfx.h"
#include "svdcmp.h"
#include "ANN/ANN.h"

// Markov Random Field Tensor Voting 
// REMARK: This function is not limited to 3D. The dimension can be set by "dim". See following.
//
// K[input/output] : The set of tensors correspond to x (You can initialize them with any tensor (NOT LIMITED TO IDENTITY))
// x               : The set of input points
// numPoint        : Number of input point
// dim             : Dimension of the point
// sigma           : Sigma of the decay function (i.e. \sigma_d of Eqn 2 (or Eqn 4 in reference [28]))
// lambda          : The weighting regarding the prior and likelihood (i.e. g of Eqn 11)
// epsBoundary     : The cutoff boundary for Eqn 2. Note that the tail of Gaussian (decay function) is long where 
//                   the values smaller than epsBoundary could be ignored.
// sor             : SOR weight (Note that we applied Gauss-Seidel Method with Successive Over-Relaxation (SOR) as the solver)
// maxIter         : Maximum number of iteration
int votingMRF( double*** K, double** x, int numPoint, int dim, double sigma, double lambda = 1.0, 
			  double epsBoundary = 0.0075 , double sor = 1.0, int maxIter = 1000000)
{
	if( K == NULL ) return 0;
	if( x == NULL ) return 0;
	if( numPoint <= 0) return 0;
	if( dim <= 0 ) return 0;
	if( sigma < EPLSION ) return 0;

	int i, j, k, m;

	double*** Kp = new double**[numPoint];
	for( i = 0; i < numPoint; i++ )
	{
		Kp[i] = allocate2DDouble( dim, dim );
		copyMatrix( K[i], Kp[i], dim, dim );
	}

	// squared neighbor radius
	double radius2 = - sigma * log( epsBoundary );

	// create neighbor list
	printf( "    [TV-MRF Msg] Create neighbor list .... ");

	ANNkd_tree* annTree = new ANNkd_tree( (ANNpointArray) x, numPoint, dim );
	int** neighborList = new int*[numPoint];
	int* neighborNum = new int[numPoint];
	double* distArray = new double[numPoint];
	int maxNumNeighbor = -MYINFINITE;
	int minNumNeighbor = MYINFINITE;
	double avgNumNeighbor = 0;
	for( i = 0; i < numPoint; i++ )
	{
		int numNeighbor = numPoint;
		numNeighbor = annTree->annkFRSearch( (ANNpoint) x[i], (ANNdist) radius2, numNeighbor );
		int* tempIndexArray = new int[numNeighbor];
		numNeighbor = annTree->annkFRSearch( (ANNpoint) x[i], (ANNdist) radius2, numNeighbor, (ANNidxArray) tempIndexArray, (ANNdistArray) distArray );
		neighborList[i] = tempIndexArray;
		neighborNum[i] = numNeighbor;
		if( numNeighbor > maxNumNeighbor ) maxNumNeighbor = numNeighbor;
		if( numNeighbor < minNumNeighbor ) minNumNeighbor = numNeighbor;
		avgNumNeighbor += double(numNeighbor);
	}
	avgNumNeighbor = avgNumNeighbor / double(numPoint);
	SAFEDEL( annTree );
	SAFEDELARR( distArray );
	annClose();
	printf( "Done\n" );

	printf("    [TV-MRF Msg] Number of neighbor (max, min, avg) = (%d, %d, %0.1f) \n", maxNumNeighbor, minNumNeighbor, avgNumNeighbor );

	bool finished = false;
	int iter = 0;
	double** newK = allocate2DDouble( dim, dim );
	double** S    = allocate2DDouble( dim, dim );
	double** dem  = allocate2DDouble( dim, dim );
	double** invD = allocate2DDouble( dim, dim );
	double** R = allocate2DDouble( dim, dim );
	double** Rp = allocate2DDouble( dim, dim );
	double** outr = allocate2DDouble( dim, dim );
	double** aux  = allocate2DDouble( dim, dim );
	double** aux2 = allocate2DDouble( dim, dim );
	double*  r = new double[ dim ];
	double** I    = allocate2DDouble( dim, dim );
	eye( dim, I );

	printf( "    [TV-MRF Msg] Voting in progress .... plsease wait .... " );

	double p = 0.95;
	double sorUpdate = false;

	if( !( (sor >= 1.0) && (sor < 2.0 - EPLSION) ) )
	{
		if( (sor > 0.0) && (sor < 1.0) )
			p = sor;
		sor = 1;
		sorUpdate = true;
	}

	while( !finished && (iter < maxIter) )
	{
		finished = true;
		
		for( m = 0; m < numPoint; m++ )
		{
			copyMatrix( K[m], newK, dim, dim );
			copyMatrix( I, dem, dim, dim );

			for( k = 0; k < neighborNum[m]; k++ )
			{	
				if( m == neighborList[m][k] ) continue;

				// ------------------- Native code for closed-form voting (Begin) --------------------
				for( i = 0; i < dim; i++ )
					r[i] = x[m][i] - x[neighborList[m][k]][i];

				double length_r2 = norm2( dim, r );
				double length_r  = sqrt( length_r2 );

				if( !nearzero( length_r ) )
					for( i = 0; i < dim; i++ )
						r[i] = r[i] / length_r;

				// outer product of r
				outerProduct( dim, r, dim, r, outr ); 

				for( i = 0; i < dim; i++ )
					for( j = 0; j < dim; j++ )
						outr[i][j] = 2.0 * outr[i][j];

				// find R
				subMat( dim, dim, I, outr, R );

				for( i = 0; i < dim; i++ )
					for( j = 0; j < dim; j++ )
						outr[i][j] = outr[i][j] / 4.0;

				// find Rp
				subMat( dim, dim, I, outr, aux );
				multiMat( dim, dim, dim, aux, R, Rp );

				//find S
				multiMat( dim, dim, dim, R, Kp[neighborList[m][k]], aux );
				multiMat( dim, dim, dim, aux, Rp, S );

				double c = exp( - fabs( length_r2 ) / sigma );

				// scale the S by distance
				for( i = 0; i < dim; i++ )
					for( j = 0; j < dim; j++ )
						S[i][j] = c * S[i][j];

				// ------------------- Native code for closed form voting (End) ----------------------

				matConstMultiply( dim, dim, S, 2.0 * lambda, S);
				addMat( dim, dim, newK, S, newK );

				multiMat( dim, dim, dim, Rp, Rp, aux );
				matConstMultiply( dim, dim, aux, lambda * c * c, aux );

				addMat( dim, dim, dem, aux, dem );

				for( i = 0; i < dim; i++ )
					dem[i][i] += lambda;

			}

			if( dim == 2 )
				matrixInv2D( dem, invD );
			else if ( dim == 3 )
				matrixInv3D( dem, invD );
			else
				inv( dem, dim, dim, invD );
			
			multiMat( dim, dim, dim, newK, invD, aux );
			copyMatrix( aux, newK, dim, dim );

			matConstMultiply( dim, dim, Kp[m], (1-sor), aux );
			matConstMultiply( dim, dim, newK, sor, aux2 );
			addMat( dim, dim, aux, aux2, newK );

			for( i = 0; i < dim; i++ )
				for( j = 0; j < dim; j++ )
					if( !nearzero( newK[i][j] - Kp[m][i][j] ) ) 
						finished = false;

			copyMatrix( newK, Kp[m], dim, dim );
		}

		if( sorUpdate )
		{
			if( iter == 0 )
				sor = 1.0 / (1.0 - p * p / 2.0);
			else
				sor = 1.0 / (1.0 - p * p * sor / 4.0);
		}

		iter++;
	}

	if( sorUpdate ) 
		printf( "    [TV-MRF Msg] Final SOR parameter = %f\n.", sor );


	// copy result to the input/output 
	for( i = 0; i < numPoint; i++ )
		copyMatrix( Kp[i], K[i], dim, dim );

	printf( "Done.\n" );

	
	// delocate temporary storage
	safeDelete2DDouble( dim, newK );
	safeDelete2DDouble( dim, S );
	safeDelete2DDouble( dim, dem );
	safeDelete2DDouble( dim, invD );
	safeDelete2DDouble( dim, R ); 
	safeDelete2DDouble( dim, Rp );
	safeDelete2DDouble( dim, outr );
	safeDelete2DDouble( dim, aux );
	safeDelete2DDouble( dim, aux2 );
	safeDelete2DDouble( dim, I );

	if( Kp != NULL )
	{
		for( i = 0; i < numPoint; i++ )
			safeDelete2DDouble( dim, Kp[i] );
		SAFEDELARR( Kp );
	}

	if( neighborList != NULL )
	{
		for( i = 0; i < numPoint; i++ )
			SAFEDELARR( neighborList[i] );
		SAFEDELARR( neighborList );
	}

	SAFEDELARR( neighborNum );

	return iter;
}

#endif

