//=======================================================================
// emtv.h : Header file of EMTV
//
// Copyright (c) 2009 Tai-Pang Wu (pang@cse.ust.hk) and HKUST x CUHK
// 
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//=======================================================================


#ifndef _EMTV_H_
#define _EMTV_H_

class EMPlane{

private:
	double** R;
	double** Rp;
	double** outr;
	double*  r;
	double** I;

	double m_fSigmaScale;
	double m_afSigma[3];
	int    m_iDim;
	int    m_iMaxIter;
	double m_fEpsBoundary;
	double m_fMEsp;

public:
	
	// Constructor 1
	// dim           : the dimension of the input
	// maxIter       : maximum number of iteration (1 iteration = one complete E- and M-Step)
	// epsBoundary   : the cutoff boundary for Eqn 2. Note that the tail of Gaussian is long where 
	//                 the values smaller than epsBoundary could be ignored.
	EMPlane( int dim, int maxIter = 1000, double epsBoundary = 0.075 );

	// Constructor 2
	// dim           : the dimension of the input
	// sigma_scale   : the scale parameter for voting (\sigma_d of Eqn 2 (or Eqn 4 in reference [28]))
	// sigma0        : the lower bound of \sigma of Eqn 5 (or Eqn 23 in reference [28])
	// sigma1        : the lower bound of \sigma1 of Eqn 5 (or Eqn 23 in reference [28])
	// sigma2        : the lower bound of \sigma2 of Eqn 5 (or Eqn 23 in reference [28])
	// maxIter       : maximum number of iteration (1 iteration = one complete E- and M-Step)
	// epsBoundary   : the cutoff boundary for Eqn 2. Note that the tail of Gaussian is long where 
	//                 the values smaller than epsBoundary could be ignored.
	EMPlane( int dim, double sigma_scale, double sigma0, double sigma1, double sigma2, int maxIter = 1000, double epsBoundary = 0.0075 );

	// Destructor
	~EMPlane();

	// EMTV on linear regression (or plane fitting)
	// x             : the set of input points
	// numPt         : number of input points
	// v[output]     : the output model (see Eqn 13 of reference [28])
	// K[output]     : the set of optimized tensors correspond to the input points x
	// w[output]     : the weights correspond to the input points x (i.e. Eqn 4 (or Eqn 19 in reference [28]))
	// return        : number of interation used to obtain the outputs
	int em_plan( double** x, int numPt, double*& v, double***& K, double*& w );

	// EMTV on F matrix estimation 
	// This function is optimized for F matrix estimation. For I/O, please see em_plan().
	int em_plan_Fmatrix( double** x, int numPt, double*& v, double***& K, double*& w );

	// It is actually tensor voting, which is called in em_plan() and em_plan_Fmatrix() to initialize the EM algorithm.
	void initialize( int numPt, double* v, double** x, double*** K, int iter );
	
	// Closed-form solution to tensor voting (Please see Eqn 1 of the full paper, or Eqn 11 of reference [28])
	void getS( double* x1, double* x2, double** K2, double** S, double sigma );	

	// Closed-form solution to tensor voting for inversed tensor (Please see Eqn 12 of reference [28])
	void getSinv( double* x1, double * x2, double** K2inv, double** Sinv, double sigma );
};

// Rank 2 enforcement for F matrix.
// For detail, plesae see R. Hartley's "In defense of the eight-point algorithm", TPAMI.
int FmatrixDiagSetZero( double* f );

// read data from file
// fileName       : file path
// numPt[output]  : number of input points
// dim[output]    : dimension of the point
double** readFromFile( const char* fileName, int& numPt, int& dim );


#endif

